@@ -286,6 +286,7 @@ def generate(specPath):
     print '# Autogenerated code by codegen.py, do not edit'
     print
     print 'import struct'
+    print 'import pika.log as log'
     print 'import pika.object'
     print 'import pika.table as table'
     print
@@ -387,6 +388,7 @@ def generate(specPath):
             if m.isSynchronous:
 
                 #Synchronous events have a CPS callback parameter
+                print "    @log.method_call"
                 print "    def %s(self, callback=None%s):" % \
                       (pyize("%s_%s" % (m.klass.name, m.name)),
                       fieldDeclList(m.arguments))
@@ -409,6 +411,7 @@ def generate(specPath):
                       ', '.join(acceptable_replies)
 
             else:
+                print "    @log.method_call"
                 print "    def %s(self%s):" % \
                       (pyize("%s_%s" % (m.klass.name, m.name)),
                       fieldDeclList(m.arguments))
